package cz.muni.fi.pa165.movie;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.util.List;
import java.util.Optional;

@Service
public class MovieService {

    private final MovieRepository repository;

    private final MovieMapper movieMapper;

    /**
     * Constructs a new instance of the MovieService class.
     *
     * @param movieRepository The repository for performing CRUD operations on Movie entities.
     * @param movieMapper The mapper for converting Movie entities to and from DTOs.
     */
    @Autowired
    public MovieService(MovieRepository movieRepository, MovieMapper movieMapper) {
        this.repository = movieRepository;
        this.movieMapper = movieMapper;
    }

    /**
     * Returns a list of all movies.
     *
     * @return the list of all movies
     */
    public List<MovieDto> getMovies () {
        return repository.findAll().stream().map(movieMapper::toDto).toList();
    }

    /**
     * Returns the movie with the specified ID.
     *
     * @param id the ID of the movie to be returned
     * @return the movie with the specified ID
     */
    public MovieDto getMovie(Long id) {
        Optional<Movie> movie = repository.findById(id);
        if (movie.isPresent()) {
            return movieMapper.toDto(movie.get());
        } else {
            throw new RuntimeException("Movie not found with id: " + id);
        }

    }

    /**
     * Creates a new movie.
     *
     * @param movieDto the movie to be created
     */
    public void createMovie(MovieDto movieDto) {
        Movie movie = movieMapper.fromDto(movieDto);
        repository.save(movie);
    }

    /**
     * Updates an existing movie.
     *
     * @param id the ID of the movie to be updated
     * @param movieDto the updated movie information
     */
    public void updateMovie(Long id, MovieDto movieDto) {
        movieMapper.updateFromDto(movieDto, repository.getReferenceById(id));
    }

    /**
     * Deletes the movie with the specified ID.
     *
     * @param id the ID of the movie to be deleted
     */
    public void deleteMovie(Long id) {
        repository.delete(repository.getReferenceById(id));
    }

    /**
     * Deletes all movies.
     */
    public void deleteAllMovies() {
        repository.deleteAll();
    }

    /**
     * Returns the movie with the specified title.
     *
     * @param title the title of the movie to be returned
     * @return the movie with the specified title
     */
    @Transactional(readOnly = true)
    public MovieDto findByTitle(String title) {
        return movieMapper.toDto(repository.findByTitle(title));
    }
}
