package cz.muni.fi.pa165.user;

import lombok.Getter;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

@Service
public class UserService {

    private final UserRepository repository;

    private final UserMapper userMapper;

    /**
     * Constructs a new UserService object.
     *
     * @param repository repository for User entities
     * @param userMapper The mapper for converting User entities to and from DTOs.
     */
    @Autowired
    public UserService(UserRepository repository, UserMapper userMapper) {
        this.repository = repository;
        this.userMapper = userMapper;
    }

    /**
     * Creates a new user.
     *
     * @param userDto the user to be created
     */
    @Transactional
    public void createUser(UserDto userDto) {
        User user = userMapper.fromDto(userDto);
        repository.save(user);
    }

    /**
     * Updates an existing user.
     *
     * @param id      the id of the user to be updated
     * @param userDto the updated user information
     */
    @Transactional
    public void updateUser(Long id, UserDto userDto) {
        userMapper.updateFromDto(userDto, repository.getReferenceById(id));
    }

    /**
     * Deletes an existing user.
     *
     * @param id the id of the user to be deleted
     */
    @Transactional
    public void deleteUser(Long id) {
        repository.delete(repository.getReferenceById(id));
    }

    /**
     * Deletes all existing users.
     */
    @Transactional
    public void deleteAllUsers() {
        repository.deleteAll();
    }

    /**
     * Returns the user that corresponds to the given username.
     *
     * @param username a String representing the username of the user to retrieve
     * @return the UserDto object that corresponds to the given username
     */
    @Transactional(readOnly = true)
    public UserDto findByUsername(String username) {
        return userMapper.toDto(repository.findByUsername(username));
    }
}
