package cz.muni.fi.pa165;

import cz.muni.fi.pa165.user.*;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.junit.jupiter.MockitoExtension;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import org.springframework.web.client.HttpClientErrorException;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.mockito.Mockito.*;

@SpringBootTest(classes = CoreApplication.class)
@AutoConfigureMockMvc
@ExtendWith({SpringExtension.class, MockitoExtension.class})
public class UserServiceTest {

    private static final String USERNAME = "Peter";
    private static final String PASSWORD = "password123456";

    @Mock
    private UserRepository userRepository;

    @InjectMocks
    private UserService userService;

    @BeforeEach
    public void setUp() {
        userService = new UserService(userRepository, new UserMapperImpl());
    }

    @Test
    public void testGetUserByUsername() {
        User user = new User();
        user.setUsername(USERNAME);
        user.setPassword(PASSWORD);

        when(userRepository.findByUsername(USERNAME)).thenReturn(user);

        UserDto result = userService.findByUsername(USERNAME);

        verify(userRepository, times(1)).findByUsername(USERNAME);
        assertThat(result).isNotNull();
        assertThat(result.getUsername()).isEqualTo(USERNAME);
        assertThat(result.getPassword()).isEqualTo(PASSWORD);
    }

    @Test
    public void testGetUserByUsernameNotFound() {
        User user = new User();
        user.setUsername(USERNAME);
        user.setPassword(PASSWORD);

        when(userRepository.findByUsername(USERNAME)).thenReturn(null);

        UserDto result = userService.findByUsername(USERNAME);

        verify(userRepository, times(1)).findByUsername(USERNAME);
        assertThat(result).isNull();
    }

    @Test
    public void testGetUserByUsernameUnauthorized() {
        User user = new User();
        user.setUsername(USERNAME);
        user.setPassword(PASSWORD);

        when(userRepository.findByUsername(USERNAME)).thenThrow(HttpClientErrorException.Unauthorized.class);

        assertThrows(HttpClientErrorException.Unauthorized.class, () -> userService.findByUsername(USERNAME));

        verify(userRepository, times(1)).findByUsername(USERNAME);
    }

    @Test
    public void testGetUserByUsernameForbidden() {
        User user = new User();
        user.setUsername(USERNAME);
        user.setPassword(PASSWORD);

        when(userRepository.findByUsername(USERNAME)).thenThrow(HttpClientErrorException.Forbidden.class);

        assertThrows(HttpClientErrorException.Forbidden.class, () -> userService.findByUsername(USERNAME));

        verify(userRepository, times(1)).findByUsername(USERNAME);
    }

    @Test
    public void testCreateUser() {
        UserDto userDto = new UserDto();
        userDto.setUsername(USERNAME);
        userDto.setPassword(PASSWORD);

        userService.createUser(userDto);

        Mockito.verify(userRepository, Mockito.times(1)).save(Mockito.any(User.class));
    }

    @Test
    public void testUpdateUser() {
        User user = new User();
        user.setUsername(USERNAME);
        user.setPassword(PASSWORD);

        UserDto userDto = new UserDto();
        userDto.setPassword("newPassword");

        when(userRepository.getReferenceById(user.getId())).thenReturn(user);

        userService.updateUser(user.getId(), userDto);

        verify(userRepository, times(1)).getReferenceById(user.getId());

        assertThat(user.getPassword()).isEqualTo(userDto.getPassword());
    }

    @Test
    public void testDeleteUser() {
        User user = new User();
        user.setUsername(USERNAME);
        user.setPassword(PASSWORD);

        when(userRepository.getReferenceById(user.getId())).thenReturn(user);

        userService.deleteUser(user.getId());

        verify(userRepository, times(1)).getReferenceById(user.getId());
        verify(userRepository, times(1)).delete(user);
    }

    @Test
    public void testDeleteAllUsers() {
        userService.deleteAllUsers();
        verify(userRepository, times(1)).deleteAll();

        assertThat(userRepository.findAll().size()).isEqualTo(0);
    }
}