package cz.muni.fi.pa165;

import cz.muni.fi.pa165.movie.GenreEnum;
import cz.muni.fi.pa165.movie.MovieDto;
import cz.muni.fi.pa165.rating.RatingDto;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.core.ParameterizedTypeReference;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.http.ResponseEntity;
import org.springframework.stereotype.Service;
import org.springframework.web.client.RestTemplate;

import java.io.*;
import java.util.List;

@Service
public class ReportService {

    @Value("${CORE_APP_URL:http://localhost:8090}")
    private String coreAppUrl;

    private final RestTemplate restTemplate;

    /**
     * Creates a new ReportService instance.
     *
     * @param restTemplate RestTemplate instance to use
     */
    @Autowired
    public ReportService(RestTemplate restTemplate) {
        this.restTemplate = restTemplate;
    }

    /**
     * Generates a report containing all movies and their average ratings.
     *
     * @param accessToken access token for authentication, can be null
     * @return byte array representing the contents of the generated file
     * @throws IOException if an I/O error occurs while generating the report
     */
    public byte[] getMoviesWithAverageRating(String accessToken) throws IOException {
        String moviesUrl = coreAppUrl + "/movies";
        String ratingsUrl = coreAppUrl + "/ratings";

        HttpEntity<?> entity = null;

        if (accessToken != null) {
            HttpHeaders headers = new HttpHeaders();
            headers.set("Authorization", "Bearer " + accessToken);
            entity = new HttpEntity<>(headers);
        }

        List<MovieDto> allMovies = getMovies(moviesUrl, entity);
        if (allMovies == null || allMovies.isEmpty()) {
            return null;
        }

        List<RatingDto> allRatings = getRatings(ratingsUrl, entity);
        if (allRatings == null || allRatings.isEmpty()) {
            return null;
        }

        return createFileWithAverageRatings(allMovies, allRatings);
    }

    /**
     * Generates a report containing movies of the specified genre and their average ratings.
     *
     * @param genre       genre to filter movies by
     * @param accessToken access token for authentication, can be null
     * @return byte array representing the contents of the generated file
     * @throws IOException if an I/O error occurs while generating the report
     */
    public byte[] getMoviesWithAverageRatingByGenre(GenreEnum genre, String accessToken) throws IOException {
        String moviesUrl = coreAppUrl + "/movies";
        String ratingsUrl = coreAppUrl + "/ratings";

        HttpEntity<?> entity = null;

        if (accessToken != null) {
            HttpHeaders headers = new HttpHeaders();
            headers.set("Authorization", "Bearer " + accessToken);
            entity = new HttpEntity<>(headers);
        }

        List<MovieDto> allMovies = getMovies(moviesUrl, entity);
        if (allMovies == null || allMovies.isEmpty()) {
            return null;
        }

        List<RatingDto> allRatings = getRatings(ratingsUrl, entity);
        if (allRatings == null || allRatings.isEmpty()) {
            return null;
        }

        List<MovieDto> moviesByGenre = allMovies.stream().filter(movie -> movie.getGenres().contains(genre)).toList();
        return createFileWithAverageRatings(moviesByGenre, allRatings);
    }

    private byte[] createFileWithAverageRatings(List<MovieDto> allMovies, List<RatingDto> allRatings) throws IOException {
        List<RatedMovie> ratedMovies = allMovies.stream()
                .map(movie -> {
                    List<RatingDto> movieRatings = allRatings.stream().filter(rating -> rating.getMovieId() == movie.getId()).toList();
                    double averageRating = calculateAverageRating(movieRatings);
                    return new RatedMovie(movie, averageRating);
                }).sorted().toList();

        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        OutputStreamWriter writer = new OutputStreamWriter(outputStream);

        for (RatedMovie ratedMovie : ratedMovies) {
            writer.write(ratedMovie.movie.getTitle() + " (" + ratedMovie.movie.getDirector() + ") \n" +
                    "   average rating: " + String.format("%.2f", ratedMovie.averageRating) + "/10\n" +
                    "   genres: " + ratedMovie.movie.getGenres() + "\n" +
                    "   actors: " + ratedMovie.movie.getActors() + "\n");
        }

        writer.flush();
        return outputStream.toByteArray();
    }

    private List<RatingDto> getRatings(String ratingsUrl, HttpEntity<?> entity) {
        ResponseEntity<List<RatingDto>> ratingsResponse = restTemplate.exchange(
                ratingsUrl,
                HttpMethod.GET,
                entity,
                new ParameterizedTypeReference<>() {
                });
        return ratingsResponse.getBody();
    }

    private List<MovieDto> getMovies(String moviesUrl, HttpEntity<?> entity) {
        ResponseEntity<List<MovieDto>> moviesResponse = restTemplate.exchange(
                moviesUrl,
                HttpMethod.GET,
                entity,
                new ParameterizedTypeReference<>() {
                });
        return moviesResponse.getBody();
    }

    private record RatedMovie(MovieDto movie, double averageRating) implements Comparable<RatedMovie> {
        @Override
        public int compareTo(RatedMovie other) {
            return Double.compare(other.averageRating, this.averageRating);
        }
    }

    private double calculateAverageRating(List<RatingDto> ratings) {
        double sumRatings = ratings.stream()
                .mapToDouble(RatingDto::getRating)
                .sum();

        if (ratings.isEmpty()) {
            return 0;
        }
        return sumRatings / ratings.size();
    }
}
