package cz.muni.fi.pa156;

import com.fasterxml.jackson.databind.ObjectMapper;
import cz.muni.fi.pa165.AppConfig;
import cz.muni.fi.pa165.ReportApplication;
import cz.muni.fi.pa165.ReportService;
import cz.muni.fi.pa165.movie.GenreEnum;
import cz.muni.fi.pa165.movie.MovieDto;
import cz.muni.fi.pa165.rating.RatingCategoryEnum;
import cz.muni.fi.pa165.rating.RatingDto;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.context.annotation.Import;
import org.springframework.http.HttpMethod;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import org.springframework.test.web.client.MockRestServiceServer;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.HttpServerErrorException;
import org.springframework.web.client.RestTemplate;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.springframework.test.web.client.match.MockRestRequestMatchers.method;
import static org.springframework.test.web.client.match.MockRestRequestMatchers.requestTo;
import static org.springframework.test.web.client.response.MockRestResponseCreators.*;

@ExtendWith(SpringExtension.class)
@SpringBootTest(classes = ReportApplication.class)
@AutoConfigureMockMvc
@Import(AppConfig.class)
@ActiveProfiles("test")
public class ReportServiceTest {

    private RestTemplate restTemplate;
    private ReportService reportService;

    @Autowired
    public ReportServiceTest(RestTemplate restTemplate, ReportService reportService) {
        this.restTemplate = restTemplate;
        this.reportService = reportService;
    }

    @Test
    public void testGetMoviesWithAverageRatingWithNoMovies() throws IOException {
        String moviesUrl = "http://localhost:8090/movies";

        MockRestServiceServer mockServer = MockRestServiceServer.createServer(restTemplate);
        mockServer.expect(requestTo(moviesUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withSuccess("[]", MediaType.APPLICATION_JSON));

        byte[] result = reportService.getMoviesWithAverageRating(null);

        assertNull(result);
        mockServer.verify();
    }

    @Test
    public void testGetMoviesWithAverageRatingWithNoMoviesUnauthorized() {
        String moviesUrl = "http://localhost:8090/movies";

        MockRestServiceServer mockServer = MockRestServiceServer.createServer(restTemplate);
        mockServer.expect(requestTo(moviesUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withUnauthorizedRequest());

        assertThrows(HttpClientErrorException.Unauthorized.class, () -> reportService.getMoviesWithAverageRating(null));

        mockServer.verify();
    }

    @Test
    public void testGetMoviesWithAverageRatingWithNoMoviesServerError() {
        String moviesUrl = "http://localhost:8090/movies";

        MockRestServiceServer mockServer = MockRestServiceServer.createServer(restTemplate);
        mockServer.expect(requestTo(moviesUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withServerError());

        assertThrows(HttpServerErrorException.class, () -> reportService.getMoviesWithAverageRating(null));

        mockServer.verify();
    }

    @Test
    public void testGetMoviesWithAverageRatingWithNoRatings() throws IOException {
        String ratingsUrl = "http://localhost:8090/ratings";
        String moviesUrl = "http://localhost:8090/movies";

        ObjectMapper objectMapper = new ObjectMapper();
        String moviesDtoJson = objectMapper.writeValueAsString(createMoviesForTesting());

        MockRestServiceServer mockServer = MockRestServiceServer.createServer(restTemplate);
        mockServer.expect(requestTo(moviesUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withSuccess(moviesDtoJson, MediaType.APPLICATION_JSON));

        mockServer.expect(requestTo(ratingsUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withSuccess("[]", MediaType.APPLICATION_JSON));

        byte[] result = reportService.getMoviesWithAverageRating(null);

        assertNull(result);
        mockServer.verify();
    }

    @Test
    public void testGetMoviesWithAverageRatingWithNoRatingsUnauthorized() throws IOException {
        String ratingsUrl = "http://localhost:8090/ratings";
        String moviesUrl = "http://localhost:8090/movies";

        ObjectMapper objectMapper = new ObjectMapper();
        String moviesDtoJson = objectMapper.writeValueAsString(createMoviesForTesting());

        MockRestServiceServer mockServer = MockRestServiceServer.createServer(restTemplate);
        mockServer.expect(requestTo(moviesUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withSuccess(moviesDtoJson, MediaType.APPLICATION_JSON));

        mockServer.expect(requestTo(ratingsUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withUnauthorizedRequest());

        assertThrows(HttpClientErrorException.Unauthorized.class, () -> reportService.getMoviesWithAverageRating(null));

        mockServer.verify();
    }

    @Test
    public void testGetMoviesWithAverageRating() throws IOException {
        String moviesUrl = "http://localhost:8090/movies";
        String ratingsUrl = "http://localhost:8090/ratings";

        ObjectMapper objectMapper = new ObjectMapper();

        String moviesDtoJson = objectMapper.writeValueAsString(createMoviesForTesting());
        String ratingsDtoJson = objectMapper.writeValueAsString(createRatingsForTesting());

        String expectedResultString = "The Godfather: Part II (Francis Ford Coppola) \n" +
                                      "   average rating: 7,00/10\n" +
                                      "   genres: [SCIFI, COMEDY]\n" +
                                      "   actors: null\n" +
                                      "The Godfather (Francis Ford Coppola) \n" +
                                      "   average rating: 3,00/10\n" +
                                      "   genres: [SCIFI, DRAMA]\n" +
                                      "   actors: null\n";

        MockRestServiceServer mockServer = MockRestServiceServer.createServer(restTemplate);
        mockServer.expect(requestTo(moviesUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withSuccess(moviesDtoJson, MediaType.APPLICATION_JSON));

        mockServer.expect(requestTo(ratingsUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withSuccess(ratingsDtoJson, MediaType.APPLICATION_JSON));

        byte[] result = reportService.getMoviesWithAverageRating( null);
        String resultString = new String(result, StandardCharsets.UTF_8);

        assertThat(resultString).isEqualTo(expectedResultString);
        mockServer.verify();
    }

    @Test
    public void testGetMoviesWithAverageRatingNoResponse() {
        String moviesUrl = "http://localhost:8090/movies";

        MockRestServiceServer mockServer = MockRestServiceServer.createServer(restTemplate);
        mockServer.expect(requestTo(moviesUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withStatus(HttpStatus.NOT_FOUND));

        assertThrows(HttpClientErrorException.NotFound.class, () -> reportService.getMoviesWithAverageRating(null));

        mockServer.verify();
    }


    @Test
    public void testGetMoviesWithAverageRatingByGenreWithNoMovies() throws IOException {
        String moviesUrl = "http://localhost:8090/movies";

        MockRestServiceServer mockServer = MockRestServiceServer.createServer(restTemplate);
        mockServer.expect(requestTo(moviesUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withSuccess("[]", MediaType.APPLICATION_JSON));

        byte[] result = reportService.getMoviesWithAverageRating(null);

        assertNull(result);
        mockServer.verify();
    }

    @Test
    public void testGetMoviesWithAverageRatingByGenreWithNoRatings() throws IOException {
        String ratingsUrl = "http://localhost:8090/ratings";
        String moviesUrl = "http://localhost:8090/movies";

        ObjectMapper objectMapper = new ObjectMapper();
        String moviesDtoJson = objectMapper.writeValueAsString(createMoviesForTesting());

        MockRestServiceServer mockServer = MockRestServiceServer.createServer(restTemplate);
        mockServer.expect(requestTo(moviesUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withSuccess(moviesDtoJson, MediaType.APPLICATION_JSON));

        mockServer.expect(requestTo(ratingsUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withSuccess("[]", MediaType.APPLICATION_JSON));

        byte[] result = reportService.getMoviesWithAverageRating(null);

        assertNull(result);
        mockServer.verify();
    }

    @Test
    public void testGetMoviesWithAverageRatingByGenre() throws IOException {
        String moviesUrl = "http://localhost:8090/movies";
        String ratingsUrl = "http://localhost:8090/ratings";

        ObjectMapper objectMapper = new ObjectMapper();
        GenreEnum testedGenre = GenreEnum.COMEDY;

        String moviesDtoJson = objectMapper.writeValueAsString(createMoviesForTesting());
        String ratingsDtoJson = objectMapper.writeValueAsString(createRatingsForTesting());

        String expectedResultString = "The Godfather: Part II (Francis Ford Coppola) \n" +
                                      "   average rating: 7,00/10\n" +
                                      "   genres: [SCIFI, COMEDY]\n" +
                                      "   actors: null\n";

        MockRestServiceServer mockServer = MockRestServiceServer.createServer(restTemplate);
        mockServer.expect(requestTo(moviesUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withSuccess(moviesDtoJson, MediaType.APPLICATION_JSON));

        mockServer.expect(requestTo(ratingsUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withSuccess(ratingsDtoJson, MediaType.APPLICATION_JSON));

        byte[] result = reportService.getMoviesWithAverageRatingByGenre(testedGenre,null);
        String resultString = new String(result, StandardCharsets.UTF_8);

        assertThat(resultString).isEqualTo(expectedResultString);
        mockServer.verify();
    }

    @Test
    public void testGetMoviesWithAverageRatingByGenreForbidden() throws IOException {
        String moviesUrl = "http://localhost:8090/movies";
        String ratingsUrl = "http://localhost:8090/ratings";

        ObjectMapper objectMapper = new ObjectMapper();
        GenreEnum testedGenre = GenreEnum.COMEDY;

        String moviesDtoJson = objectMapper.writeValueAsString(createMoviesForTesting());

        MockRestServiceServer mockServer = MockRestServiceServer.createServer(restTemplate);
        mockServer.expect(requestTo(moviesUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withSuccess(moviesDtoJson, MediaType.APPLICATION_JSON));

        mockServer.expect(requestTo(ratingsUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withForbiddenRequest());

        assertThrows(HttpClientErrorException.Forbidden.class, () -> reportService.getMoviesWithAverageRatingByGenre(testedGenre, null));

        mockServer.verify();
    }


    private List<MovieDto> createMoviesForTesting() {
        List<MovieDto> movies = new ArrayList<>();

        MovieDto movie1 = new MovieDto();
        movie1.setId(1L);
        movie1.setTitle("The Godfather");
        movie1.setDescription("A classic crime drama");
        movie1.setDirector("Francis Ford Coppola");
        movie1.setGenres(Arrays.asList(GenreEnum.SCIFI, GenreEnum.DRAMA));
        movies.add(movie1);

        MovieDto movie2 = new MovieDto();
        movie2.setId(2L);
        movie2.setTitle("The Godfather: Part II");
        movie2.setDescription("A sequel to the classic crime drama");
        movie2.setDirector("Francis Ford Coppola");
        movie2.setGenres(Arrays.asList(GenreEnum.SCIFI, GenreEnum.COMEDY));
        movies.add(movie2);

        return movies;
    }

    private List<RatingDto> createRatingsForTesting() {
        List<RatingDto> ratings = new ArrayList<>();

        RatingDto rating1 = new RatingDto();
        rating1.setId(1L);
        rating1.setCategory(RatingCategoryEnum.PLOT);
        rating1.setRating(2);
        rating1.setUserId(1L);
        rating1.setMovieId(1L);
        ratings.add(rating1);

        RatingDto rating2 = new RatingDto();
        rating2.setId(2L);
        rating2.setCategory(RatingCategoryEnum.CAST);
        rating2.setRating(4);
        rating2.setUserId(1L);
        rating2.setMovieId(1L);
        ratings.add(rating2);

        RatingDto rating3 = new RatingDto();
        rating3.setId(3L);
        rating3.setCategory(RatingCategoryEnum.PLOT);
        rating3.setRating(8);
        rating3.setUserId(1L);
        rating3.setMovieId(2L);
        ratings.add(rating3);

        RatingDto rating4 = new RatingDto();
        rating4.setId(4L);
        rating4.setCategory(RatingCategoryEnum.CAST);
        rating4.setRating(6);
        rating4.setUserId(1L);
        rating4.setMovieId(2L);
        ratings.add(rating4);

        return ratings;
    }
}
